/*
 * 15:41:50 06/04/99
 *
 * JextTextArea.java - An extended JTextArea
 * Copyright (C) 1998-1999 Romain Guy
 * powerteam@chez.com
 * www.chez.com/powerteam
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.chez.powerteam.jext;

import java.io.*;
import java.awt.*;
import gnu.regexp.*;
import java.awt.event.*;
import javax.swing.JFrame;
import javax.swing.undo.*;
import javax.swing.event.*;
import javax.swing.JTextArea;
import javax.swing.JOptionPane;
import javax.swing.text.Element;
import com.chez.powerteam.jext.menus.Indent;
import javax.swing.text.BadLocationException;

public class JextTextArea extends JTextArea implements CaretListener, UndoableEditListener, KeyListener, DocumentListener
{
  // JextTextArea extends JTextArea but doesn't override it
  // Implements Caret, Undo, Document and Key Listeners

  // Private declaration

  private RE regexp;
  private Jext parent;
  private CompoundEdit compoundEdit;
  private String fontName, currentFile;
  private boolean dirty, newf, operation;
  private int anchor, fontSize, fontStyle;
  private UndoManager undo = new UndoManager();

  // Constants

  private static final int BUFFER_SIZE = 32768;

  // Must have a Jext parent

  JextTextArea(Jext parent)
  {
    super();

    addKeyListener(this);
    addCaretListener(this);
    getDocument().addUndoableEditListener(this);
    getDocument().addDocumentListener(this);

    setBorder(null);
    setLineWrap(false);
    setCaretColor(Color.red);
    setCursor(new Cursor(Cursor.TEXT_CURSOR));

    this.parent = parent;
    Font defaultFont = new Font("Monospaced", Font.PLAIN, 12);
    fontName = defaultFont.getName();
    fontSize = defaultFont.getSize();
    fontStyle = defaultFont.getStyle();
    setFont(defaultFont);

    newf = true;
  }

  // Get infos about the current font

  public String getFontName()
  {
    return fontName;
  }

  public int getFontSize()
  {
    return fontSize;
  }

  public int getFontStyle()
  {
    return fontStyle;
  }

  // Set the font to be used

  public void setFontName(String name)
  {
    fontName = name;
    changeFont();
  }

  public void setFontSize(int size)
  {
    fontSize = size;
    changeFont();
    repaint();
  }

  public void setFontStyle(int style)
  {
    fontStyle = style;
    changeFont();
    repaint();
  }

  private void changeFont()
  {
    setFont(new Font(fontName, fontStyle, fontSize));
  }

  // Called when the setChanged() method musn't be called

  public void beginOperation()
  {
    operation = true;
  }

  // Called to allow the DocumentListener to call the
  // setChanged() method

  public void endOperation()
  {
    operation = false;
  }

  // Return parent of this

  public Jext getJextParent()
  {
    return parent;
  }

  // Determines if we can call the setChanged() method

  public boolean getOperation()
  {
    return operation;
  }

  // What is the current opened file ?

  public String getCurrentFile()
  {
    return currentFile;
  }

  // Clean the area

  public void newDocument()
  {
    beginOperation();
    if (isDirty() && !isEmpty())
    {
      int response = JOptionPane.showConfirmDialog(null, "Do you want to save your changes ?", "Save", JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
      switch (response)
      {
        case 0:
          saveContent();
	  break;
	case 1:
	  break;
        case 2:
          return;
	default:
	  return;
      }
    }
    clean();
    newf = true;
    currentFile = null;
    setEditable(true);
    setText(null);
    setAnchor(0);
    discard();
    parent.setNew();
    parent.setTitle("Jext - Java Text Editor");
    endOperation();
  }

  // See AutoSave.java

  public void autoSave()
  {
    if (!isNew()) saveContent();
  }

  // Called to save this component's content
  // Call save(String file) but let the user choose a file name
  // if the isNew() flag is true (int the case the user choosed
  // an existing file, we ask him if he really want to overwrite it)

  public void saveContent()
  {
    if (!isEditable()) return;
    if (isNew())
    {
      String fileToLoad = Utilities.choosePlace(parent, Utilities.SAVE);
      if (fileToLoad != null)
      {
        File shouldBeSaved = new File(fileToLoad);
        if (!shouldBeSaved.exists())
          save(fileToLoad);
	else
	{
          int response = JOptionPane.showConfirmDialog(null, "File already exists, overwrite it ?", "Save", JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
          switch (response)
          {
            case 0:
              save(fileToLoad);
              break;
            case 1:
              break;
            default:
              return;
          }
	}
      }
    } else {
      if (isDirty()) save(currentFile);
    }
  }

  // Called to save this component's content

  public void save(String file)
  {
    try
    {
      OutputStream outs = new FileOutputStream(new File(file));
      BufferedWriter out = new BufferedWriter(new OutputStreamWriter(outs), BUFFER_SIZE);
      String newline = System.getProperty("line.separator");
      Element map = getDocument().getDefaultRootElement();
      for (int i = 0; i < map.getElementCount(); i++)
      {
        Element line = map.getElement(i);
        int start = line.getStartOffset();
        out.write(getText(start, line.getEndOffset() - start - 1));
        out.write(newline);
      }
      out.close();
      outs.close();
      if (!file.equals(currentFile)) currentFile = file;
      if (isNew()) newf = false;
      parent.setTitle("Jext - " + file);
      parent.setSaved();
    } catch (IOException ioe) { Utilities.showError("Error has occured while saving"); }
    catch (BadLocationException ble) { Utilities.showError("Error has occured while saving"); }
  }

  // Called to load a new file in the text area
  // Determines which line separator (\n, \r\n...) are used in the
  // file to open. Convert'em into Swing line separator (\n).

  public void load(String path)
  {
    clean();
    discard();
    beginOperation();
    InputStream is;
    StringBuffer buffer = new StringBuffer();
    try
    {
      File toLoad = new File(path);
      if (!toLoad.canWrite())
        setEditable(false);
      else if (!isEditable())
        setEditable(true);
      is = new FileInputStream(toLoad);
      InputStreamReader in = new InputStreamReader(is);
      char[] buf = new char[BUFFER_SIZE];
      int len;
      int lineCount = 0;
      boolean CRLF = false;
      boolean CROnly = false;
      boolean lastWasCR = false;

      while ((len = in.read(buf, 0, buf.length)) != -1)
      {
        int lastLine = 0;
        for (int i = 0; i < len; i++)
        {
          switch(buf[i])
          {
            case '\r':
              if (lastWasCR)
              {
                CROnly = true;
                CRLF = false;
              } else
                lastWasCR = true;
              buffer.append(buf, lastLine, i - lastLine);
              buffer.append('\n');
              lastLine = i + 1;
              break;
            case '\n':
              if (lastWasCR)
              {
                CROnly = false;
                CRLF = true;
                lastWasCR = false;
                lastLine = i + 1;
              } else {
                CROnly = false;
                CRLF = false;
                buffer.append(buf, lastLine, i - lastLine);
                buffer.append('\n');
                lastLine = i + 1;
              }
              break;
            default:
              if (lastWasCR)
              {
                CROnly = true;
                CRLF = false;
                lastWasCR = false;
              }
              break;
          }
        }
        buffer.append(buf, lastLine, len - lastLine);
      }
      in.close();
      if (buffer.length() != 0 && buffer.charAt(buffer.length() - 1) == '\n')
        buffer.setLength(buffer.length() - 1);

      getDocument().remove(0, getLength());
      getDocument().insertString(0, buffer.toString(), null);
      currentFile = path;
      setCaretPosition(0);
      parent.setNew();
      newf = false;
      parent.setTitle("Jext - " + Utilities.getShortStringOf(path, 80));
      parent.saveRecent(path);
    } catch(BadLocationException bl) {
      bl.printStackTrace();
    } catch(FileNotFoundException fnf) {
      Utilities.showError(path + " not found !");
    } catch(IOException io) {
      Utilities.showError(io.toString());
    }
    endOperation();
  }

  // Return true if current text is new

  public boolean isNew()
  {
    return newf;
  }

  // Return true if area is empty

  public boolean isEmpty()
  {
    if (getLength() == 0)
      return true;
    else
      return false;
  }

  // Return true if area content has changed, false otherwise

  public boolean isDirty()
  {
    return dirty;
  }

  // Called when the content of the area has changed

  public void setDirty()
  {
    dirty = true;
  }

  // Called after having saved or created a new document to ensure
  // the content isn't 'dirty'

  public void clean()
  {
    dirty = false;
  }

  // Turn tabs into 'soft tabs' (tab to spaces)

  private void softTab(KeyEvent evt)
  {
    if (!parent.getSoftTab()) return;
    try
    {
      getDocument().insertString(getCaretPosition(), Utilities.createWhiteSpace(getTabSize()), null);
      evt.consume();
    } catch (BadLocationException ble) { }
  }

  // Discard all edits contained in the UndoManager
  // Update the corresponding menu items

  public void discard()
  {
    undo.discardAllEdits();
    parent.discard();
  }

  // Return the anchor position

  public int getAnchor()
  {
    return anchor;
  }

  // Set the anchor postion to offset

  public void setAnchor(int offset)
  {
    anchor = offset;
  }

  // Used by Jext to update its menu items

  public UndoManager getUndo()
  {
    return undo;
  }

  // Used for ReplaceAll
  // This merges all text changes made between the beginCompoundEdit()
  // and the endCompoundEdit() calls into only one undo event

  public void beginCompoundEdit()
  {
    if (compoundEdit == null)
      compoundEdit = new CompoundEdit();
  }

  public void endCompoundEdit()
  {
    if (compoundEdit != null)
    {
      compoundEdit.end();
      if (compoundEdit.canUndo())
        undo.addEdit(compoundEdit);
      compoundEdit = null;
      parent.undoUpdate();
      parent.redoUpdate();
    }
  }

  // Return the lentgh of the text in the area

  public int getLength()
  {
    return getDocument().getLength();
  }

  // Return true if searchStr has been found, false otherwise

  public boolean find(String searchStr, int start, boolean ignoreCase)
  {
    try
    {
      if (searchStr.equals("") || searchStr == null) return false;
      regexp = new RE(searchStr, (ignoreCase == true ? RE.REG_ICASE : 0) | RE.REG_MULTILINE, RESyntax.RE_SYNTAX_PERL5);
      if (regexp == null)
      {
        getToolkit().beep();
        return false;
      }
      String text = getText(start, getLength() - start);
      REMatch match = regexp.getMatch(text);
      if (match != null)
      {
        this.select(start + match.getStartIndex(), start + match.getEndIndex());
        return true;
      }
    } catch(Exception e) { }
    return false;
  }

  // Return true if the replace has been succesfully done, false in any
  // other case

  public boolean replaceAll(String searchStr, String replaceStr, int start, int end, boolean ignoreCase) throws REException
  {
    boolean found = false;
    beginCompoundEdit();
    try
    {
      if (searchStr.equals("") || searchStr == null) return false;
      regexp = new RE(searchStr, (ignoreCase == true ? RE.REG_ICASE : 0) | RE.REG_MULTILINE, RESyntax.RE_SYNTAX_PERL5);
      if (regexp == null)
      {
        endCompoundEdit();
        return false;
      }

      if (replaceStr == null)
        replaceStr = "";

      REMatch match;

      Element map = getDocument().getDefaultRootElement();
      int startLine = map.getElementIndex(start);
      int endLine = map.getElementIndex(end);

      for (int i = startLine; i <= endLine; i++)
      {
        Element lineElement = map.getElement(i);
        int lineStart;
        int lineEnd;

        if (i == startLine)
          lineStart = start;
        else
          lineStart = lineElement.getStartOffset();
        if (i == endLine)
          lineEnd = end;
        else
          lineEnd = lineElement.getEndOffset() - 1;

        lineEnd -= lineStart;
        String line = getText(lineStart, lineEnd);
        String newLine = regexp.substituteAll(line, replaceStr);
        if (line.equals(newLine)) continue;
        getDocument().remove(lineStart, lineEnd);
        getDocument().insertString(lineStart, newLine, null);

        end += (newLine.length() - lineEnd);
        found = true;
      }
    } catch(Exception e) {
      found = false;
    }
    endCompoundEdit();
    return found;
  }

  // Abstract method of the CaretListener

  public void caretUpdate(CaretEvent evt)
  {
    parent.updateStatus(evt.getDot());
  }

  // Abstract method of the UndoableEditListener

  public void undoableEditHappened(UndoableEditEvent e)
  {
    if (!getOperation())
    {
      if (compoundEdit == null)
      {
        undo.addEdit(e.getEdit());
        parent.undoUpdate();
        parent.redoUpdate();
      } else
        compoundEdit.addEdit(e.getEdit());
    }
  }

  // Abstract methods of the KeyListener

  public void keyPressed(KeyEvent evt)
  {
    if (evt.getKeyChar() == KeyEvent.VK_ENTER && parent.getIndent())
    {
      try
      {
        getDocument().insertString(getCaretPosition(), "\n", null);
      } catch (BadLocationException ble) { }
      if (Indent.indent())
        evt.consume();
      else
        evt.consume();
    } else if (evt.getKeyChar() == KeyEvent.VK_TAB) {
      if (parent.getIndent())
      {
        if (Indent.indent())
	  evt.consume();
	else
	  softTab(evt);
      } else if (parent.getSoftTab())
        softTab(evt);
    } else if (evt.getKeyChar() == KeyEvent.VK_DELETE)
      parent.updateStatus(getCaretPosition());
  }

  public void keyReleased(KeyEvent evt) { }
  public void keyTyped(KeyEvent evt) { }

  // Abstract methods of the DocumentListener

  public void changedUpdate(DocumentEvent e)
  {
    if (!getOperation()) parent.setChanged();
  }

  public void insertUpdate(DocumentEvent e)
  {
    if (!getOperation()) parent.setChanged();
  }

  public void removeUpdate(DocumentEvent e)
  {
    if (!getOperation()) parent.setChanged();
  }
}

// End of JextTextArea.java
